#
# FetchFile.pm: library for fetching files locally or via FTP
# Copyright (C) 1998 Roman Hodek <Roman.Hodek@informatik.uni-erlangen.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Id: FetchFile.pm,v 1.22 2001/01/15 10:47:54 rnhodek Exp $
#
# $Log: FetchFile.pm,v $
# Revision 1.22  2001/01/15 10:47:54  rnhodek
# Search less dirs for FTP pool structures.
# Init FTP pool struct correctly.
#
# Revision 1.21  2001/01/11 14:42:23  rnhodek
# Search less dirs under pool/ (use index).
# Implement %dist_aliases (to work around strange unstable symlink).
#
# Revision 1.20  2000/12/20 18:55:47  rnhodek
# Additionally scan pool/ subdirectories for packages.
#
# Revision 1.19  2000/01/25 16:20:38  rnhodek
# Fix typo in FF_init (cache wasn't cleared properly).
#
# Revision 1.18  1999/09/03 13:49:58  rnhodek
# After reading sbuild config, also delete $INC{/etc/sbuild.conf.local}.
#
# Revision 1.17  1999/08/26 08:55:02  rnhodek
# fetch_ftp: use only one $local_file var.
# fetch_ftp(lftp): fix typo.
#
# Revision 1.16  1999/08/10 13:48:12  rnhodek
# Redirect error output of ftpget (lftp code).
#
# Revision 1.15  1999/07/30 09:07:39  rnhodek
# ftp_code() and ftp_response() needs $chan as argument.
#
# Revision 1.14  1999/07/26 06:03:13  rnhodek
# Fix typo.
#
# Revision 1.13  1999/07/23 16:24:52  rnhodek
# Print "download aborted... retrying" only if the file existed already.
#
# Revision 1.12  1999/07/21 15:14:12  rnhodek
# fetch_ftp: print FTP result code and text response also if code isn't
# 2xx (i.e, not successful).
#
# Revision 1.11  1999/07/21 15:00:15  rnhodek
# Make fetch_ftp() (via Net::FTP) more fault-tolerant: It retries to
# download the file as long as the local size is smaller than the
# expected size. The retrying is aborted on FTP errors and if a
# downloaded didn't increase the local file's size. Finally also too big
# files are rejected.
# In case of FTP errors, also print the last error code and the response
# string.
#
# Revision 1.10  1999/05/28 12:10:34  rnhodek
# Remove unused fetch_ftp_via_lftp function.
# Make $now in fetch_ftp local to lftp part.
# In lftp command in ftp_cache_add use && as separator instead of ';' to
# avoid listing the root dir if $dir doesn't exist.
#
# Revision 1.9  1999/05/28 09:01:00  rnhodek
# Fixed lftp things: In fetch_ftp and $root path too much was added;
# lftp dir command can append /s to names, remove them.
#
# Revision 1.8  1999/03/05 13:37:26  rnhodek
# Redesign of cache: Do not cache names of all files anymore, but only
# directory names. Full caching is somewhat wasteful and longish, and
# the cache has to be cleared rather often after "not found"s.
#
# Revision 1.7  1999/02/23 16:49:55  rnhodek
# Handle URLs as file names and use wget to fetch such things.
#
# Revision 1.6  1998/11/26 04:48:22  james
# lftp preserves timestamps on files, which confuses buildd-watcher, so touch them post-download.
#
# Revision 1.5  1998/11/23 02:32:16  james
# Support for (optionally) using lftp in FetchFile.
#
# Revision 1.4  1998/10/31 02:19:10  james
# Comment typo fixes. [Courtesy of the English Pedantry association.]
#
# Revision 1.3  1998/10/26 15:00:07  rnhodek
# Get option $ftp_timeout from conf file and pass it on to Net::FTP->new.
#
# Revision 1.2  1998/10/08 11:33:33  rnhodek
# Special handling for Packages files in init_cache_ftp; they're at a
# level where usually on dirs are.
#
# Revision 1.1  1998/10/08 11:21:04  rnhodek
# Initial writing.
#
#

package FetchFile;

# read sbuild config files if not done already
if (!defined @conf::archive_locations) {
	package sbuild_conf;
	require "/etc/sbuild.conf" if -r "/etc/sbuild.conf";
	delete $INC{"/etc/sbuild.conf"};
	delete $INC{"/etc/sbuild.conf.local"};
	require "$main::HOME/.sbuildrc" if -r "$main::HOME/.sbuildrc";
	delete $INC{"$main::HOME/.sbuildrc"};
	package FetchFile;
	@FetchFile::archive_locations = @sbuild_conf::archive_locations;
	@FetchFile::dist_parts = @sbuild_conf::dist_parts;
	%FetchFile::dist_aliases = %sbuild_conf::dist_aliases;
	$FetchFile::ftp_timeout = $sbuild_conf::ftp_timeout;
	$FetchFile::use_lftp = $sbuild_conf::use_lftp;
}
else {
	@FetchFile::archive_locations = @conf::archive_locations;
	@FetchFile::dist_parts = @conf::dist_parts;
	%FetchFile::dist_aliases = %conf::dist_aliases;
	$FetchFile::ftp_timeout = $conf::ftp_timeout;
	$Fetchfile::use_lftp = $conf::use_lftp;
}

$FetchFile::warner = \&default_warner;
$FetchFile::ftp_timeout ||= 120;
$FetchFile::use_lftp ||= 0;

use strict;
use Net::FTP;
use POSIX;

require Exporter;
@FetchFile::ISA = qw(Exporter);
@FetchFile::EXPORT = qw(FF_init FF_search FF_search_different_version FF_fetch);

sub w(@);

sub FF_init {
	my $dist = shift;
	my $srcbin = shift || "source";

	# invalidate caches if necessary
	if ((defined $FetchFile::dist && $FetchFile::dist ne $dist) ||
		(defined $FetchFile::srcbin && $FetchFile::srcbin ne $srcbin)) {
		undef %FetchFile::cache;
	}
	
	$FetchFile::dist = $FetchFile::dist_aliases{$dist} || $dist;
	$FetchFile::srcbin = $srcbin;
}

sub init_cache {
	my $loc = shift;
	my $debug = shift;
	my @args = (\@{$FetchFile::cache{$loc}}, $loc, $debug);

	print "Initializing cache for $loc\n" if $debug;
	if (is_local($loc)) {
		return init_cache_local(@args);
	}
	elsif (is_ftp($loc)) {
		return init_cache_ftp(@args);
	}
	else {
		w "cannot determine type of location $loc\n";
	}
	return 0;
}

sub init_cache_local {
	my $cacheref = shift;
	my $root = shift;
	my $debug = shift;
	my(@paths, $dir, $dpath);
	local( *PIPE );

	@paths = map { "$root/dists/$FetchFile::dist/$_/$FetchFile::srcbin" }
				 @FetchFile::dist_parts;
	push( @paths, map { "$root/pool/$_" } @FetchFile::dist_parts );
	push( @paths, "$root/dists/proposed-updates" )
		if $FetchFile::dist eq 'stable';
	foreach $dir (@paths) {
		next if ! -d $dir;
		push( @{$cacheref}, $dir );
		print "Looking for dirs in $dir\n" if $debug;
		if (!opendir( DIR, $dir )) {
			w "Can't open dir $dir: $!\n";
			return 0;
		}
		my $de;
		while( $de = readdir(DIR) ) {
			next if $de =~ /^\.\.?$/;
			$dpath = "$dir/$de";
			my $filesfound = 0;
			if (-d $dpath) {
				if (opendir( SDIR, $dpath )) {
					my ($de2, $dpath2);
					while( $de2 = readdir(SDIR) ) {
						next if $de2 =~ /^\.\.?$/;
						$dpath2 = "$dpath/$de2";
						if (-d $dpath2) {
							push( @{$cacheref}, $dpath2 );
						}
						elsif (-f _) {
							$filesfound = 1;
						}
					}
					closedir( SDIR );
			  	}
				push( @{$cacheref}, $dpath ) if $filesfound;
			}
		}
		closedir( DIR );
	}
	return 1;
}

sub init_cache_ftp {
	my $cacheref = shift;
	my $loc = shift;
	my $debug = shift;
	my($host, $root) = ftp_parts($loc);
	my($part, $chan);
	
	$chan = open_ftp( $host, $root, $debug ) or return 0
		if !$FetchFile::use_lftp;
	
	foreach $part (@FetchFile::dist_parts) {
		ftp_cache_add( $cacheref, $chan, $host, $debug,
					   "$root/dists/$FetchFile::dist/$part/$FetchFile::srcbin");
		ftp_cache_add( $cacheref, $chan, $host, $debug,
					   "$root/pool/$part");
	}
	if ($FetchFile::dist eq 'stable') {
		ftp_cache_add( $cacheref, $chan, $host, $debug,
					   "$root/dists/proposed-updates" );
	}
	return 1;
}

sub ftp_cache_add {
	my $cacheref = shift;
	my $chan = shift;
	my $host = shift;
	my $debug = shift;
	my $dir = shift;
	
	$chan->cwd( $dir )
		or (w "can't cd to $dir on $host\n", return)
			if !$FetchFile::use_lftp;
	print "Looking for dirs in ftp://$host$dir\n" if $debug;

	my @subdirs = $FetchFile::use_lftp ?
		split ("\n", `lftp -e "cd $dir && dir; quit" $host 2> /dev/null`) :
		@{$chan->dir()};
	w "can't list $dir on $host\n", return if !@subdirs;

	my $filesfound = 0;
	foreach (@subdirs) {
		my @f = split /\s+/;
		my $mode = $f[0];
		my $name = $f[$#f];
		$name =~ s,/+$,,;
		next if $name =~ /^\.\.?$/;
		if ($mode =~ /^d/) {
			ftp_cache_add( $cacheref, $chan, $host, $debug, "$dir/$name" );
			push( @{$cacheref}, "$dir/$name" );
		} else {
			$filesfound = 1;
		}
	}
	push( @{$cacheref}, "$dir" ) if $filesfound;
}


sub FF_search {
	my $file = shift;
	my $debug = shift;
	my $loc;
	
	print "Searching for $file\n" if $debug;
	return $file if is_url($file);
	
	foreach $loc (@FetchFile::archive_locations) {
		if (!$FetchFile::cache{$loc}) {
			next if !init_cache( $loc, $debug );
		}
		my $dir;
		foreach $dir (@{$FetchFile::cache{$loc}}) {
			my $found = 0;
			if ($dir =~ m,/pool/,) {
				my $index = substr( $file, 0, ($file =~ /^lib/) ? 4 : 1 );
				next if $dir !~ m,/$index/,; 
			}
			if (is_local($loc)) {
				print "Testing $dir/$file\n" if $debug;
				$found = 1 if -f "$dir/$file";
			}
			elsif (is_ftp($loc)) {
				my($host, $root) = ftp_parts($loc);
				print "Testing ftp://$host$dir/$file\n" if $debug;
				if (!$FetchFile::use_lftp) {
					my $chan = open_ftp( $host, $root, $debug );
					next if !$chan;
					$found = 1 if $chan->ls( "$dir/$file" );
				}
				else {
					$found = 1 if `lftp -e "nlist $dir/$file; quit" $host 2> /dev/null`;
				}
			}
			else {
				w "cannot determine type of location $loc\n";
			}
			if ($found) {
				print "Found as $dir/$file in location $loc\n" if $debug;
				return "$dir/$file";
			}
		}
		print "Not found in $loc\n" if $debug;
	}
	print "...not found\n" if $debug;
	return "";
}

sub FF_search_different_version {
	my $file = shift;
	my $debug = shift;
	my $loc;
	
	return "" if $file !~ /^([^_]+)_([^_]+)\.dsc$/ || is_url($file);
	my $pattern = $1."_*.dsc";

	foreach $loc (@FetchFile::archive_locations) {
		if (!$FetchFile::cache{$loc}) {
			next if !init_cache( $loc, $debug );
		}
		my $dir;
		foreach $dir (@{$FetchFile::cache{$loc}}) {
			my $found;
			if (is_local($loc)) {
				print "Testing $dir/$pattern\n" if $debug;
				my @files = glob( "$dir/$pattern" );
				$found = $files[0] if @files;
			}
			elsif (is_ftp($loc)) {
				my($host, $root) = ftp_parts($loc);
				print "Testing ftp://$host/$dir/$pattern\n" if $debug;
				if (!$FetchFile::use_lftp) {
					my $chan = open_ftp( $host, $root, $debug );
					next if !$chan;
					my @files = $chan->ls( "$dir/$pattern" );
					$found = $files[0] if @files && defined $files[0];
				}
				else {
					my @files = split( "\n", `lftp -e "nlist $dir/$file; quit" $host 2> /dev/null` );
					$found = $files[0] if @files;
				}
			}
			else {
				w "cannot determine type of location $loc\n";
			}
			if ($found) {
				return basename($found);
			}
		}
	}
	return "";
}

sub FF_fetch {
	my $file = shift;
	my $debug = shift;
	my $loc;
	my $first = 1;

	print "Fetching file $file\n" if $debug;
	if (is_url($file)) {
		if (system( "wget -q $file" ) != 0) {
			w "wget of $file failed\n";
			return 0;
		}
		return 1;
	}
	
	foreach $loc (@FetchFile::archive_locations) {
		w "Trying next source: $loc\n" if !$first;
		$first = 0;
		if (is_local($loc)) {
			# for fetching a local file, just make a symlink
			if (!-f "$file") {
				w "$file: not found\n";
				next;
			}
			my $base = basename( $file );
			if (!symlink( "$file", $base )) {
				w "Can't create symlink $base: $!\n";
				next;
			}
			return 1;
		}
		elsif (is_ftp($loc)) {
			return 1 if fetch_ftp( $loc, $file, $debug );
		}
		else {
			w "cannot determine type of location $loc\n";
		}
	}
	return 0;
}

sub fetch_ftp {
	my $loc = shift;
	my $file = shift;
	my $debug = shift;
	my $chan;
	my($host,$root) = ftp_parts($loc);
	my $local_file = basename($file);
	
	if (!$FetchFile::use_lftp) {
		$chan = open_ftp( $host, $root, $debug ) or return 0;
		my $wanted_size = $chan->size( $file );
		if (!defined $wanted_size) {
			w "$loc/$file: can't get size (probably file doesn't exist)\n";
			return 0;
		}
		for(;;) {
			my $current_size;
			$current_size = -s _ if -f $local_file;
			last if defined($current_size) && $current_size >= $wanted_size;
			w "$loc/$file: download aborted at $current_size; retrying\n"
				if defined($current_size);
			if (!$chan->get( $file, $local_file, $current_size )) {
				w "$loc/$file: retrieval failed\n";
				my ($code, $resp) = (ftp_code($chan), ftp_response($chan));
				w "FTP response $code $resp\n";
				unlink( $local_file );
				return 0;
			}
			my $code = ftp_code($chan);
			if ($code !~ /^2/) {
				my $resp = ftp_response($chan);
				w "$loc/$file: FTP error\n";
				w "FTP response $code $resp\n";
			}
			my $new_size = -s $local_file;
			if ($new_size <= $current_size) {
				w "$loc/$file: retry failed (nothing downloaded)\n";
				unlink( $local_file );
				return 0;
			}
		}
		my $local_size = -s $local_file;
		if ($local_size != $wanted_size) {
			w "$loc/$file: download error, file has wrong size $local_size (expected $wanted_size)\n";
			unlink( $local_file );
			return 0;
		}
	} else {
		system "ftpget -q $host $file 2> /dev/null";
		if ($?) {
			w "$loc/$file: retrieval failed\n";
			return 0;
		}
		# lftp preserves the timestamps on downloaded files,
		# confusing the hell out of buildd-watcher; so we
		# `fix' the timestamps.
		my $now = time;
		utime $now, $now, $local_file;
	}

	return 1;
}

sub ftp_response() {
	my $chan = shift;
	return join( '', @{${*$chan}{'net_cmd_resp'}} );
}

sub ftp_code() {
	my $chan = shift;
	return ${*$chan}{'net_cmd_code'};
}

sub is_local {
	return ($_[0] =~ m,^/,) ? 1 : 0;
}

sub is_ftp {
	return ($_[0] =~ m,^ftp://,) ? 1 : 0;
}

sub ftp_parts {
	$_[0] =~ m,^ftp://([^/]*)(/.*)$,;
	return ($1, $2);
}

sub open_ftp {
	my $host = shift;
	my $dir = shift;
	my $debug = shift;
	my $chan;

	if ($chan = $FetchFile::ftp_open{$host}) {
		return $chan if $chan->cwd( $dir );
		print "cwd on old FTP connection failed: $!\n" if $debug;
		# probably connection timed out
		print "Reopening...\n" if $debug;
		$FetchFile::ftp_open{$host} = undef;
	}

	if (!($chan = Net::FTP->new( $host, Debug => $debug ? $debug-1 : 0,
								 Timeout => $FetchFile::ftp_timeout ))) {
		w "Cannot open FTP server $host\n";
		return 0;
	}
	if (!$chan->login()) {
		w "Anonymous login to $host failed\n";
		return 0;
	}
	if (!$chan->binary()) {
		w "Can't set binary FTP mode on $host\n";
		return 0;
	}
	if (!$chan->cwd( $dir )) {
		w "Can't cd to $dir on $host\n";
		return 0;
	}
	$FetchFile::ftp_open{$host} = $chan;
	return $chan;
}

sub basename {
	my $b = $_[0];
	$b =~ s,^.*/,,;
	return $b;
}

sub is_url {
	my $f = shift;
	return ($f =~ m,^\w+://,) ? 1 : 0;
}

sub w(@) {
	&{$FetchFile::warner}( "FetchFile: ", @_ );
}

sub default_warner {
	warn @_;
}

1;
